/*
 * calmwm - the calm window manager
 *
 * Copyright (c) 2004 Marius Aamodt Eriksen <marius@monkey.org>
 * All rights reserved.
 *
 * $Id: cycle.c,v 1.5 2004/07/10 20:50:13 marius Exp $
 */

#include "headers.h"
#include "calmwm.h"

struct cycle *
cycle_new(void)
{
	struct cycle *cy;

	XCALLOC(cy);
	TAILQ_INIT(&cy->cycleq);
	TAILQ_INIT(&cy->disabledq);

	return (cy);
}

void
cycle_free(struct cycle *cy)
{
	struct cycle_entry *cye;

	while ((cye = TAILQ_FIRST(&cy->cycleq)) != NULL) {
		TAILQ_REMOVE(&cy->cycleq, cye, entry);
		xfree(cye);
	}

	xfree(cy);
}

void
cycle_add(struct cycle *cy, void *ctx)
{
	struct cycle_entry *cye;

	XMALLOC(cye);
	cye->ctx = ctx;
	TAILQ_INSERT_TAIL(&cy->cycleq, cye, entry);

	if (cy->current == NULL)
		cy->current = cye;
}

/*
 * XXX this is ugly.  We should either hash ctx -> entries, or make
 * the user of the API keep entries.
 */
void
cycle_remove(struct cycle *cy, void *ctx)
{
	struct cycle_entry *cye, *nextcye;

	for (cye = TAILQ_FIRST(&cy->cycleq); cye != NULL; cye = nextcye) {
		nextcye = TAILQ_NEXT(cye, entry);

		if (cye->ctx == ctx) {
			if (cy->current == cye)
				cy->current = NULL;
			if (cy->begin == cye)
				cy->begin = NULL;

			TAILQ_REMOVE(&cy->cycleq, cye, entry);
			xfree(cye);
		}
	}
}

void
cycle_disable(struct cycle *cy, void *ctx)
{
	struct cycle_entry *cye, *nextcye;

	for (cye = TAILQ_FIRST(&cy->cycleq); cye != NULL; cye = nextcye) {
		nextcye = TAILQ_NEXT(cye, entry);

		if (cye->ctx == ctx) {
			TAILQ_REMOVE(&cy->cycleq, cye, entry);
			TAILQ_INSERT_TAIL(&cy->disabledq, cye, entry);
		}

		if (cy->current == cye)
			cy->current = NULL;
		if (cy->begin == cye)
			cy->begin = NULL;
	}

}

void
cycle_enable(struct cycle *cy, void *ctx)
{
	struct cycle_entry *cye, *nextcye;

	for (cye = TAILQ_FIRST(&cy->disabledq); cye != NULL; cye = nextcye) {
		nextcye = TAILQ_NEXT(cye, entry);

		if (cye->ctx == ctx) {
			TAILQ_REMOVE(&cy->disabledq, cye, entry);
			TAILQ_INSERT_TAIL(&cy->cycleq, cye, entry);
		}
	}
}

void *
cycle_next(struct cycle *cy, int backward, int begin)
{
	struct cycle_entry *cye;

	if (begin) {
		cye = cy->begin;
		cy->begin = cy->current;
		cy->current = cye;
	} else {
		if (cy->current != NULL) {
			cye = cy->current;
			cy->current = backward ?
			    TAILQ_PREV(cye, cycle_entry_q, entry) :
			    TAILQ_NEXT(cye, entry);
		}
	}

	if (cy->current == NULL)
		cy->current = TAILQ_FIRST(&cy->cycleq);

	return (cy->current != NULL ? cy->current->ctx : NULL);
}

void
cycle_current(struct cycle *cy, void *ctx)
{
	struct cycle_entry *cye;

	TAILQ_FOREACH(cye, &cy->cycleq, entry)
		if (cye->ctx == ctx)
			break;

	if (cye == NULL)
		return;

	cy->begin = cy->current;
	cy->current = cye;
}
